/*--------------------------------------------------------------------------
@COPYRIGHT  :
              Copyright 1996, Alex P. Zijdenbos, 
              McConnell Brain Imaging Centre,
              Montreal Neurological Institute, McGill University.
              Permission to use, copy, modify, and distribute this
              software and its documentation for any purpose and without
              fee is hereby granted, provided that the above copyright
              notice appear in all copies.  The author and McGill University
              make no representations about the suitability of this
              software for any purpose.  It is provided "as is" without
              express or implied warranty.
---------------------------------------------------------------------------- 
$RCSfile: CachedArray.h,v $
$Revision: 1.3 $
$Author: alex $
$Date: 1997/10/03 17:44:59 $
$State: Exp $
--------------------------------------------------------------------------*/
#ifndef _CACHED_ARRAY_H
#define _CACHED_ARRAY_H

#include "SimpleArray.h"
#include <fstream.h>

/********************************************************************
 * CachedArray class
 *
 * This class is derived from SimpleArray, and reduces memory use
 * by caching to disk.
 ********************************************************************/

template <class Type> class CachedArray;
template <class Type> class CacheBlock;

/*
typedef CachedArray<char>          CachedBoolArray;
typedef CachedArray<int>           CachedIntArray;
typedef CachedArray<unsigned>      CachedUnsignedArray;
typedef CachedArray<double>        CachedDblArray;
*/

template <class Type>
class CachedArray : public SimpleArray<Type> {
  static const unsigned _DEFAULT_BLOCK_SIZE;
  static const unsigned _DEFAULT_N_BLOCKS;

  CachedArray<Type> *_self;       // Provided to circumvent the const mechanism
  CacheBlock<Type>  *_head;       // Pointer to first cache block in linked list
  CacheBlock<Type> **_blocks;     // Pointers to cache blocks
  unsigned           _blockSize;  // # elements in each cache block
  unsigned           _nBlocks;    // Current # of cache blocks
  unsigned           _maxNblocks; // Maximum # cache blocks (= length of block table)
  fstream            _s;          // The stream associated with the temporary file
  // Hitrate calculation members
  unsigned           _hits;       // Cache hit counter
  unsigned           _misses;     // Cache miss counter
  // Iterator members
  unsigned           _itBlock;
  Type              *_itBlockPtr;

public:
  // Manager functions
  CachedArray(unsigned sz = 0, unsigned nBlocks = _DEFAULT_N_BLOCKS,
	      unsigned blockSize = _DEFAULT_BLOCK_SIZE);
  CachedArray(const Type *init, unsigned sz, unsigned nBlocks = _DEFAULT_N_BLOCKS,
	      unsigned blockSize = _DEFAULT_BLOCK_SIZE);
  CachedArray(const CachedArray<Type>& array);
  CachedArray(const SimpleArray<Type>& array, unsigned nBlocks = _DEFAULT_N_BLOCKS,
	      unsigned blockSize = _DEFAULT_BLOCK_SIZE);
  virtual ~CachedArray();

  // Binary I/O functions
  ostream& saveBinary(ostream&, unsigned n = 0, unsigned start = 0) const;
  istream& loadBinary(istream&, unsigned n = 0, unsigned start = 0);

  // Matlab I/O functions
#ifdef HAVE_MATLAB
  virtual Boolean saveMatlab(const char *, const char *, const char *) const {
    _notImplementedError(); return FALSE; }
#endif

  // Access functions
  Type& getEl(unsigned i)              { return _block(i)->getEl(i % _blockSize); }
  const Type& getEl(unsigned i) const  { return getElConst(i); }
  const Type& getElConst(unsigned i) const { 
    return _block(i)->getElConst(i % _blockSize); }
  void  setEl(unsigned i, Type value)  { _block(i)->setEl(i % _blockSize, value); }

  // Iterator functions
  //void  resetIterator(unsigned i = 0);       // Resets iterator to element i
  void  resetIterator(unsigned i = 0) const;
  Type& current();
  const Type& current() const;

  // Prefix ascending iterators
  Type& operator ++();
  const Type&  operator ++() const;
  // Postfix ascending iterators
  Type& operator ++(int);
  const Type&  operator ++(int) const;

  // Prefix descending iterators
  Type& operator --();
  const Type&  operator --() const;
  // Postfix descending iterators
  Type& operator --(int);
  const Type&  operator --(int) const;

  // Copy operators
  CachedArray& operator = (const CachedArray<Type>& array) {
    Array<Type>::operator = (array); return *this; }
  CachedArray& copyFromCarray(const Type *init, unsigned sz);

  // Conversions
  //operator SimpleArray<Type> () const;

  // Set functions
  void     newSize(unsigned sz);     // Resize array
  void     resetHitRate() const { 
    CachedArray<Type> *self = (CachedArray<Type>*) this; self->_hits=self->_misses=0; }

  // Not implemented
  SimpleArray<Type> operator () (unsigned) const {
    _notImplementedError(); return SimpleArray<Type>(0); }
  SimpleArray<Type> operator () (unsigned, unsigned) const {
    _notImplementedError(); return SimpleArray<Type>(0); }
  SimpleArray<Type> operator () (const BoolArray&) const {
    _notImplementedError(); return SimpleArray<Type>(0); }
  SimpleArray<Type> operator () (const UnsignedArray&) const {
    _notImplementedError(); return SimpleArray<Type>(0); }
  SimpleArray<Type> common(const SimpleArray<Type>&) const {
    _notImplementedError(); return SimpleArray<Type>(0); }

  Type& first() { return getEl(0); }
  Type& last()  { return getEl(_size - 1); }
  const Type    *contents() const  { _notImplementedError(); return 0; }
  Type          *contents()        { _notImplementedError(); return 0; }
  operator const Type *() const    { _notImplementedError(); return 0; }
  operator       Type *()          { _notImplementedError(); return 0; }

  // Quicksort functions; not implemented
  void qsort() { _notImplementedError(); }
  void qsort(int (*) (const void *, const void *)) { _notImplementedError(); }
  void qsortAscending() { _notImplementedError(); }
  void qsortDescending() { _notImplementedError(); }
  UnsignedArray qsortIndexAscending() const { 
    _notImplementedError(); return UnsignedArray(0); }
  UnsignedArray qsortIndexDescending() const {
    _notImplementedError(); return UnsignedArray(0); }

  // Get functions
  double hitRate() const;

  // Median functions. The array is first reduced, using a histogram approach,
  // to a single-block array, for which the algorithm described
  // in Sections 8.1, 8.3, and 10.2 in Cormen, Leierson, and Rivest,
  // "Introduction to Algorithms" (aka the Big White Book) is used.
  // The low median is returned in case the array has an even # elements.
  Type median() const;
  Type medianVolatile();

  CachedArray sample(unsigned maxN) const;
  CachedArray applyElementWise(Type (*function) (Type)) const;
  CachedArray map(const ValueMap& map) const;

  CachedArray ln() const;                    // Natural logarithm of all elements
  CachedArray log() const;                   // Base-10 logarithm of all elements
  CachedArray exp() const;                   // e^(all elements)
  CachedArray exp10() const;                 // 10^(all elements)

private:
  void _initialize(unsigned sz, unsigned nBlocks = _DEFAULT_N_BLOCKS,
		   unsigned blockSize = _DEFAULT_BLOCK_SIZE);
  void              _openStream();
  void              _destroy();
  CacheBlock<Type> *_block(unsigned i) const {
    if (i >= _size)
      _rangeError(i);
    CacheBlock<Type> *block = _blocks[i / _blockSize];
    return (block) ? block : _read(i / _blockSize);
  }
  void              _flush() const;              // Flush cache to disk
  CacheBlock<Type> *_read(unsigned block) const; // Reads block i
  void              _write() const;              // Write current block
  void              _revertIterator() const {    // Revert iterator (load correct block)
    _self->_itBlockPtr = _read(_self->_itBlock)->_contents; }
    

  // Median support functions
  Type _histMedian(unsigned nBelow = 0, unsigned nAbove = 0);
  Type _randomizedSelect(int p, int r, int i);
  int  _randomizedPartition(int p, int r);
  int  _partition(int p, int r);
};

template <class Type>
CachedArray<Type> operator ^ (double base, const CachedArray<Type>& array);

/********************************************************************
 * CacheBlock class
 *
 * Support class for CachedArray. 
 ********************************************************************/

template <class Type>
class CacheBlock : private SimpleArray<Type> {
  friend class CachedArray<Type>;

  CacheBlock<Type> *_self;

  CacheBlock *_next;
  unsigned    _nBytes;
  Boolean     _changed;
  unsigned    _ID;
  unsigned    _nRead;
  unsigned    _nWrite;

public:  
  CacheBlock(unsigned ID, unsigned sz);
  ~CacheBlock();
  
  Type&       getEl(unsigned i) { _changed = TRUE; _nWrite++; return _contents[i]; }
  const Type& getElConst(unsigned i) const  { _self->_nRead++; return _contents[i]; }
  void        setEl(unsigned i, Type value) { 
    _changed = TRUE; _nWrite++; _contents[i]=value; }

  //operator SimpleArray<Type> () const { return SimpleArray<Type>(_contents, _size); }

  CacheBlock *addBlock(unsigned ID, unsigned sz);
  Boolean     read(fstream& s, unsigned ID); // Read new block; flushes existing one 
  Boolean     write(fstream&) const;
};

#endif
